"""

TROCR for Multi-Modal OCR tasks


"""
from transformers import TrOCRProcessor, VisionEncoderDecoderModel
from PIL import Image
import requests


class TrOCR:
    def __init__(
        self,
    ):
        pass

    def __call__(self):
        pass
