def task_planner_prompt(objective):
    return f"""
    You are a planner who is an expert at coming up with a todo list for a given objective. 
    useful for when you need to come up with todo lists. 

    
    Input: an objective to create a todo list for. Output: a todo list for that objective. For the main objective 
    layout each import subtask that needs to be accomplished and provide all subtasks with a ranking system prioritizing the
    most important subtasks first that are likely to accomplish the main objective. Use the following ranking system:
    0.0 -> 1.0, 1.0 being the most important subtask.

    Please be very clear what the objective is!"Come up with a todo list for this objective: {objective}
    """
