from swarms.prompts.code_interpreter import CODE_INTERPRETER
from swarms.prompts.finance_agent_prompt import FINANCE_AGENT_PROMPT
from swarms.prompts.growth_agent_prompt import GROWTH_AGENT_PROMPT
from swarms.prompts.legal_agent_prompt import LEGAL_AGENT_PROMPT
from swarms.prompts.operations_agent_prompt import OPERATIONS_AGENT_PROMPT
from swarms.prompts.product_agent_prompt import PRODUCT_AGENT_PROMPT
